/* $Id$

(c) Copyright, Real-Time Innovations, 2022.
All rights reserved.

No duplications, whole or partial, manual or electronic, may be made
without express written permission.  Any such copies, or
revisions thereof, must display this notice unaltered.
This code contains trade secrets of Real-Time Innovations, Inc.
*/

#ifndef RTI_DDS_TOPIC_TRUST_PARTICIPANT_TRUST_ALGORITHM_INFO_HPP_
#define RTI_DDS_TOPIC_TRUST_PARTICIPANT_TRUST_ALGORITHM_INFO_HPP_

// IMPORTANT: macros.hpp must be the first RTI header included in every header
// file so that symbols are exported correctly on Windows
#include <dds/core/macros.hpp>

#include "dds_c/dds_c_infrastructure_impl.h"

#include <dds/core/types.hpp>
#include <dds/core/Value.hpp>
#include <dds/core/Exception.hpp>
#include <rti/core/NativeValueType.hpp>
#include <rti/topic/trust/ParticipantTrustSignatureAlgorithmInfo.hpp>
#include <rti/topic/trust/ParticipantTrustKeyEstablishmentAlgorithmInfo.hpp>
#include <rti/topic/trust/ParticipantTrustInterceptorAlgorithmInfo.hpp>

namespace rti { namespace topic { namespace trust {

class ParticipantTrustAlgorithmInfoAdapter {
public:
    typedef struct DDS_ParticipantTrustAlgorithmInfo native_type;

    static void initialize(native_type& native_value)
    {
        static const native_type DEFAULT =
                DDS_ParticipantTrustAlgorithmInfo_INITIALIZER_NOSEC;
        native_value = DEFAULT;
    }

    static void finalize(native_type&)
    {
        // nothing to do
    }

    static void copy(native_type& destination, const native_type& source)
    {
        DDS_ParticipantTrustAlgorithmInfo_copy(&destination, &source);
    }

    static bool equals(const native_type& first, const native_type& second)
    {
        return DDS_ParticipantTrustAlgorithmInfo_equals(&first, &second)
                  == DDS_BOOLEAN_TRUE;
    }
};

class ParticipantTrustAlgorithmInfo;

} } // namespace topic::trust

namespace core {

template <>
struct native_type_traits<rti::topic::trust::ParticipantTrustAlgorithmInfo> {
    typedef rti::topic::trust::ParticipantTrustAlgorithmInfoAdapter adapter_type;
    typedef DDS_ParticipantTrustAlgorithmInfo native_type;
};

}

namespace topic { namespace trust {

/**
 * @ingroup DDSBuiltInTopicCommonModule
 * @RTI_class_definition class ParticipantTrustAlgorithmInfo
 * @headerfile rti/topic/trust/ParticipantTrustAlgorithmInfo.hpp
 *
 * @brief @extension @st_value_type Trust Plugins algorithm information
 * associated with the discovered DomainParticipant.
 *
 * @details \dref_details_ParticipantTrustAlgorithmInfo
 */
class OMG_DDS_API ParticipantTrustAlgorithmInfo : public rti::core::NativeValueType<ParticipantTrustAlgorithmInfo> {
public:
    RTI_NATIVE_VALUE_TYPE_DEFINE_DEFAULT_MOVE_OPERATIONS(ParticipantTrustAlgorithmInfo)

    typedef rti::core::NativeValueType<ParticipantTrustAlgorithmInfo> Base;
    /**
     * @brief Create an instance with the default Trust Algorithm Info
     * associated with the discovered DomainParticipant.
     *
     * @details \dref_details_ParticipantTrustAlgorithmInfo_new
     */
    ParticipantTrustAlgorithmInfo() = default;

    ParticipantTrustAlgorithmInfo(
            const struct DDS_ParticipantTrustAlgorithmInfo& native_algorithms)
      : Base(native_algorithms)
    {
    }

    /**
     * @brief Get the Trust Plugins algorithms for validation of data and
     * metadata exchanged by the DomainParticipant.
     *
     * @details \dref_details_ParticipantTrustAlgorithmInfo_signature
     */
    ParticipantTrustSignatureAlgorithmInfo signature() const
    {
        return native().signature;
    }

    /**
     * @brief Get the Trust Plugins algorithms for transformation of data and
     * metadata exchanged by the DomainParticipant.
     *
     * @details \dref_details_ParticipantTrustAlgorithmInfo_key_establishment
     */
    ParticipantTrustKeyEstablishmentAlgorithmInfo key_establishment() const
    {
        return native().key_establishment;
    }

    /**
     * @brief Get the Trust Plugins algorithms for interception of data and
     * metadata exchanged by the DomainParticipant.
     *
     * @details \dref_details_ParticipantTrustAlgorithmInfo_interceptor
     */
    ParticipantTrustInterceptorAlgorithmInfo interceptor() const
    {
        return native().interceptor;
    }

};

} } }

#endif // RTI_DDS_TOPIC_TRUST_PARTICIPANT_TRUST_ALGORITHM_INFO_HPP_
