/* $Id$

 (c) Copyright, Real-Time Innovations, 2021
 All rights reserved.
 No duplications, whole or partial, manual or electronic, may be made
 without express written permission.  Any such copies, or
 revisions thereof, must display this notice unaltered.
 This code contains trade secrets of Real-Time Innovations, Inc.

=========================================================================*/

#ifndef monitoring_monitoringClass_h
#define monitoring_monitoringClass_h

#ifndef monitoring_dll_h
  #include "monitoring/monitoring_dll.h"
#endif
#ifndef ndds_c_h
  #include "ndds/ndds_c.h"
#endif
#ifndef dds_c_trust_plugins_h
  #include "dds_c/dds_c_trust_plugins.h"
#endif
#ifndef dds_c_monitoring_h
  #include "dds_c/dds_c_monitoring.h"
#endif
#ifndef reda_inlineMemory_h
  #include "reda/reda_inlineMemory.h"
#endif

#ifdef __cplusplus
    extern "C" {
#endif

/*i
 * @brief This function prototype allows creating a distribution metric group
 * object from the equivalent application metric group object.
 *
 * The function prototype uses the input memory manager to allocate memory for
 * the new distribution metric group object.
 *
 * @param appMetricGroup In. Application metric group object. Cannot be NULL.
 * @param memoryManager In. Memory Manager from which to allocate the 
 * distribution metric group object. We have to provide a double pointer because
 * the memory manager can be reallocated.
 *
 * @return distribution metric group object or NULL is there is an error.
 */
typedef void * (*RTI_Monitoring_MapApplicationToDistributionMetricGroupFunction)(
      const void *appMetricGroup,
      struct REDAInlineMemory **memoryManager);

/*i
 * @brief This function deletes a distribution metric group object allocated 
 * from a memory manager using mapAppToDistMetricGroupFunction.
 *
 * @param distMetricGroup InOut. Distribution metric group object. 
 * Cannot be NULL.
 */
typedef void (*RTI_Monitoring_DeleteDistributionMetricGroupFunction)(
      void *distMetricGroup);

/*i
 * @brief This function prototype allows assigning a distribution metric group
 * object to the sample from which it will be published.
 *
 * The assignment does not make a copy because the distribution metric group
 * is an optional member in the distribution sample.
 */
typedef void (*RTI_Monitoring_SetDistributionMetricGroupIntoSampleFunction)(
      void *sample,
      void *distMetricGroup);

/*i
 * @brief This structure contains all the information required to manipulate
 * a metric group.
 *
 * For each structure marked with the @resource annotation the Code Generator
 * generates a function called <Resource Type>_get_metric_group_info() that 
 * returns an array of RTI_Monitoring_MetricGroupInfo objects. There is one
 * element per metric group associated with the resource.
 *
 * The elements in the RTI_Monitoring_MetricGroupInfo can be accessed using the
 * index constants that are part of the distribution model generated with the
 * option -convertToMonitor.
 *
 * For example: DATA_WRITER_STATUS_PROTOCOL_METRIC_GROUP_INDEX
 */
struct RTI_Monitoring_MetricGroupInfo {
    /*i
     * @brief The metric group ID. This ID can be also accessed through the
     * generated constants in the distribution IDL file.
     */
    DDS_UnsignedLong groupId;
    /*i
     * @brief This function allows creating a distribution metric group object
     * from the equivalent application metric group object.
     */
    RTI_Monitoring_MapApplicationToDistributionMetricGroupFunction
            mapAppToDistMetricGroupFunction;
    /*i
     * @brief This function allows assigning a distribution metric group object
     * to the sample from which it will be published.
     */
    RTI_Monitoring_SetDistributionMetricGroupIntoSampleFunction
            setDistMetricGroupIntoSample;
    /*i
     * @brief This function deletes a distribution metric group object allocated
     * from a memory manager using mapAppToDistMetricGroupFunction.
     */
    RTI_Monitoring_DeleteDistributionMetricGroupFunction
            deleteDistMetricGroupFunction;
};

#define RTI_Monitoring_MetricGroupInfo_INITIALIZER { \
    0, /* groupId */ \
    NULL, /* mapAppToDistMetricGroupFunction */ \
    NULL, /* setDistMetricGroupIntoSample */ \
    NULL /* deleteDistMetricGroupFunction */ \
}

/*i
 * @brief This structure contains information about an individual metric in
 * a metric group.
 */
struct RTI_Monitoring_MetricInfo {
    /*i
     * @brief The metric name. This is the name used to store metrics in
     * Observability backends. The name uses '.' to separate the different
     * components. We also refer to this name as the OpenTelemetry name.
     */
    char *name;
    /*i
     * @brief The metric ID. This ID can be also accessed through the
     * generated constants in the distribution IDL file.
     */
    DDS_UnsignedLong metricId;
    /*i
     * @brief The Index of the metric group containing this metric. If there are
     * multiple metric groups containing this metric, this index corresponds
     * to the metric group with the highest observable level.
     *
     * The index refers to the position in the array of metric groups returned
     * by the <Resource Type>_get_metric_group_info() function.
     */
    DDS_UnsignedLong metricGroupIndex;

    /*i
     * @brief The distribution kind of the metric.
     */
    DDS_ObservableDistributionKind distributionKind;
};

#define RTI_Monitoring_MetricInfo_INITIALIZER { \
    NULL, /* name */ \
    0, /* metricId */ \
    0, /* metricGroupIndex */ \
    DDS_OBSERVABLE_UNSPECIFIED_DISTRIBUTION_KIND /* distributionKind */ \
}

/*i
 * @brief This structure contains all the information required to manipulate
 * a resource which is code generation specific. 
 * 
 * For example, the number of periodic metric groups and event metric groups.
 * 
 * This information is needed to keep the implementation of the monitoring
 * library agnostic to specific resources and metric groups defined in the
 * data model.
 */
struct RTI_Monitoring_ResourceInfo {
    /*i @brief The number of periodic metric groups in the resource. */
    DDS_UnsignedLong periodicMetricGroupCount;
    /*i @brief The number of event metric groups in the resource. */
    DDS_UnsignedLong eventMetricGroupCount;
    /*i
     * @brief The number of metrics in the resource. This number includes the
     * metrics in the periodic and event metric groups.
     */
    DDS_UnsignedLong metricCount;
};

#define RTI_Monitoring_ResourceInfo_INITIALIZER { \
    0, /* periodicMetricGroupCount */ \
    0, /* eventMetricGroupCount */ \
    0 /* metricCount */ \
}

/*i
 * @brief Create an instance of RTI_MonitoringClass.
 * @param factory \b In. The DDS_DomainParticipantFactory used by the
 * RTI_MonitoringClass instance to create the DDS entities used to publish
 * telemetry data.
 * @param property \b In. The property for the Monitoring Library.
 * @return A Monitoring class instance or NULL if there is an error.
 */
extern RTI_MonitoringDllExport
struct RTI_MonitoringClass * RTI_Monitoring_create(
        DDS_DomainParticipantFactory *factory,
        const struct RTI_MonitoringProperty_t *property);

/*i
 * @brief Pre-finalize the Monitoring Class instance.
 * @details This method should perform any finalization operation that needs to
 * be done before taking any concurrency model mutex.
 * @param monitoring_class \b InOut. The Monitoring class instance to
 * pre-finalize.
 * @return DDS_BOOLEAN_TRUE if the instance was pre-finalized successfully.
 * DDS_BOOLEAN_FALSE otherwise.
 */
extern RTI_MonitoringDllExport DDS_Boolean
RTI_Monitoring_prefinalize(struct RTI_MonitoringClass *monitoring_class);

/*i
 * @brief Delete the Monitoring Library class instance.
 * @param monitoring_class \b In. The Monitoring class instance to remove.
 */
extern RTI_MonitoringDllExport
void RTI_Monitoring_delete(struct RTI_MonitoringClass *monitoring_class);

/*i
 * @brief Initialize the create_function pointer of the Monitoring Library.
 * This function is called by the user if linking and by the application
 * automatically if loading dynamically.
 */
extern RTI_MonitoringDllExport
void RTI_Monitoring_initialize(void);

/*i
 * @brief Check if the Monitoring Library is compiled dynamically or statically.
 * @return DDS_BOOLEAN_TRUE if dynamically and DDS_BOOLEAN_FALSE if statically.
 */
extern RTI_MonitoringDllExport
DDS_Boolean RTI_Monitoring_is_shared_library(void);

/*
 * Observability Framework Category Labels.
 * These defines represent the category labels that will be sent in the
 * structured_data field of DDS_Monitoring_Logging. Some of them map to
 * Connext Logging Categories or to Security Plugin Class families.
 */
#define RTI_MONITORING_LOGGING_DISCOVERY_CATEGORY_LABEL \
  RTI_LOG_CATEGORY_DISCOVERY_LABEL
#define RTI_MONITORING_LOGGING_SECURITY_CATEGORY_LABEL \
  RTI_LOG_CATEGORY_SECURITY_LABEL
#define RTI_MONITORING_LOGGING_NOT_APPLICABLE_CATEGORY_LABEL "N/A"

/* User-friendly names for security plugin class names */
#define RTI_MONITORING_LOGGING_SEC_AUTHENTICATION_CATEGORY_LABEL \
  DDS_LOGGING_AUTHENTICATION_CLASS_FAMILY
#define RTI_MONITORING_LOGGING_SEC_ACCESS_CONTROL_CATEGORY_LABEL \
  DDS_LOGGING_ACCESS_CONTROL_CLASS_FAMILY
#define RTI_MONITORING_LOGGING_SEC_INTERCEPTOR_CATEGORY_LABEL \
  DDS_LOGGING_INTERCEPTOR_CLASS_FAMILY
#define RTI_MONITORING_LOGGING_SEC_LOGGING_CATEGORY_LABEL \
  DDS_LOGGING_LOGGING_CLASS_FAMILY

/* Separator for categories */
#define RTI_MONITORING_LOGGING_CATEGORY_LABEL_SEPARATOR ","

/* Structured data key names */
#define RTI_MONITORING_LOGGING_CONNEXT_KEY "Connext"

/* Structured data name value pairs names */
#define RTI_MONITORING_LOGGING_CATEGORY_LABEL_NAME "category"

#ifdef __cplusplus
    }	/* extern "C" */
#endif

#endif /* monitoring_monitoringClass_h */
/* end of $Id$ */