/*
 * (c) Copyright, Real-Time Innovations, 2002-2022.
 * All rights reserved.
 * No duplications, whole or partial, manual or electronic, may be made
 * without express written permission.  Any such copies, or
 * revisions thereof, must display this notice unaltered.
 * This code contains trade secrets of Real-Time Innovations, Inc.
 */

#ifndef rtitest_setting_impl_h
#define rtitest_setting_impl_h

#include "rtitest/test_dll.h"
#include "rtitest/test_log.h"
#include "rtitest/test_connext_interface.h"

#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

/**************************** Global configuration ****************************/

/**
 * @ingroup RTITestSetting
 * @brief Does some setup for the Unit Test Framework. Must be called in the 
 * top-level Tester run method of a test executable.  
 */
#define RTITestSetting_setupWithConnext() \
  RTITestSetting_initializeSetup(); \
  RTITestConnext_setup()

/** @internal
 * @ingroup RTITestSetting
 * @brief THIS SHOULD NOT BE USED FOR CONNEXT TESTING. Configures test module
 * without setting up the state required to run Connext tests.
 */
#define RTITestSetting_setupStandalone() RTITestSetting_initializeSetup()

/****************************** Default settings ******************************/

#if defined(RTI_VXWORKS) && !defined(RTI_RTP)
  #define RTI_TEST_SETTING_MAX_RETRIES_DEFAULT 1
#else
  #define RTI_TEST_SETTING_MAX_RETRIES_DEFAULT 0
#endif

#define RTI_TEST_SETTING_DEFAULT \
  { \
    { -1, -1 }, /* begin */ \
    { -1, -1 }, /* end */ \
    RTI_TEST_LOG_BITMAP_DEFAULT, \
    /* instrumentationMask */ \
    0,                                    /* ignoreFailure */ \
    0,                                    /* runInteractiveTest */ \
    0,                                    /* id */ \
    NULL,                                 /* tag */ \
    0,                                    /* activeSync */ \
    0,                                    /* testOdbcDatabases */ \
    0,                                    /* keepOdbcResources */ \
    NULL,                                 /* odbcIniPath */ \
    NULL,                                 /* execPath */ \
    "",                                   /* database */ \
    0,                                    /* showTime */ \
    "",                                   /* user */ \
    "",                                   /* _jUnitOutputFile */ \
    NULL,                                 /* _jUnitOutputFD */ \
    0,                                    /* _level */ \
    RTI_TEST_SETTING_MAX_RETRIES_DEFAULT, /* _maxRetries */ \
    0,                                    /* noRtiLan */ \
    0,                                    /* noTestFiles */ \
    0,                                    /* noWritableFileSystem */ \
    0,                                    /* repeatTests */ \
    0,                                    /* numberOfRepeats */ \
    0,    /* globalIndividualUnitTestRetryCount */ \
    NULL, /* option */ \
    0,    /* enableHeapMonitoring */ \
    0,    /* enableNetworkCapture */ \
    0,    /* valgrindPerTest */ \
    0,    /* detectInterferingParticipants */ \
    RTI_FALSE,     /* enableLeakSanitizerCheck */ \
    RTI_FALSE     /* failOnLeak */ \
  }

#define RTITestSetting_set( \
        setting, \
        beginModule, \
        beginUnit, \
        endModule, \
        endUnit, \
        instMask, \
        ignore, \
        uniqueId, \
        runInteractive, \
        junitLogFile) \
  { \
    (setting)->begin.submoduleTestNumber = (beginModule); \
    (setting)->begin.unitTestNumber = (beginUnit); \
    (setting)->end.submoduleTestNumber = (endModule); \
    (setting)->end.unitTestNumber = (endUnit); \
    (setting)->instrumentationMask = (instMask); \
    (setting)->ignoreFailure = (ignore); \
    (setting)->id = (uniqueId); \
    (setting)->runInteractiveTest = runInteractive; \
    strncpy((setting)->_jUnitOutputFile, \
            junitLogFile, \
            RTI_TEST_MAX_FILE_NAME_LENGTH); \
    (setting)->_jUnitOutputFile[RTI_TEST_MAX_FILE_NAME_LENGTH] = '\0'; \
    (setting)->noRtiLan = 0; \
    (setting)->noTestFiles = 0; \
    (setting)->noWritableFileSystem = 0; \
  }

/*************************** Heap monitoring support **************************/

extern RTITestDllExport RTIBool RTITestSetting_setHeapMonitoringSupport(void);

#define RTITestSetting_enableHeapMonitoringSupport() \
  RTIOsapiHeap_enableMonitoring() ? RTITestSetting_setHeapMonitoringSupport() \
                                  : RTI_FALSE

extern RTITestDllExport void RTITestSetting_unsetHeapMonitoringSupport(void);

#define RTITestSetting_disableHeapMonitoringSupport(__ok) \
  { \
    struct RTIOsapiHeapStat stats__ = RTI_OSAPI_HEAP_STAT_DEFAULT; \
    RTIOsapiHeap_statGet(&stats__); \
    if (stats__.currentUsage != 0) { \
      *(__ok) = 0; \
      RTITestLog_logParamString( \
              RTI_TEST_LOG_BIT_EXCEPTION, \
              "%s: detected memory leak. %lld bytes are still allocated\n", \
              RTI_FUNCTION_NAME, \
              stats__.currentUsage); \
    } \
    RTIOsapiHeap_disableMonitoring(); \
    RTITestSetting_unsetHeapMonitoringSupport(); \
  }

/*************************** Network capture support **************************/

extern RTITestDllExport RTIBool RTITestSetting_setNetworkCaptureSupport(void);

/**
 * @internal
 *
 * @brief Start network capture for all DomainParticipants.
 *
 * @param[out] ok It will be set to RTI_FALSE if starting network capture fails.
 */
#define RTITestSetting_startNetworkCapture(ok__) \
{ \
    NDDS_Utility_NetworkCaptureParams_t networkCaptureParams__ = \
            NDDS_UTILITY_NETWORK_CAPTURE_PARAMETERS_DEFAULT; \
    networkCaptureParams__.parse_encrypted_content = DDS_BOOLEAN_TRUE; \
    *(ok__) = NDDS_Utility_start_network_capture_w_params( \
            RTITestSetting_g_networkCaptureFileNamePrefix, \
            &networkCaptureParams__); \
    NDDS_Utility_NetworkCaptureParams_t_finalize(&networkCaptureParams__); \
}

#define RTITestSetting_enableAndStartNetworkCaptureSupport(ok__) \
{ \
    RTIBool isNetworkCaptureEnabled__ = NDDS_Utility_enable_network_capture(); \
    if (!isNetworkCaptureEnabled__) { \
        *(ok__) = RTI_FALSE; \
        RTITestLogger_log( \
                RTI_TEST_LOG_BIT_EXCEPTION, \
                &RTI_TEST_LOG_FAILURE_TEMPLATE, \
                "Error enabling network capture."); \
    } \
    RTITestSetting_startNetworkCapture(ok__); \
    if (!*(ok__)) { \
        RTITestLogger_log( \
                RTI_TEST_LOG_BIT_EXCEPTION, \
                &RTI_TEST_LOG_FAILURE_TEMPLATE, \
                "Error starting network capture."); \
    } \
    RTITestSetting_setNetworkCaptureSupport(); \
}

extern RTITestDllExport void RTITestSetting_unsetNetworkCaptureSupport(void);

#define RTITestSetting_stopAndDisableNetworkCaptureSupport(ok__) \
  { \
    if (!NDDS_Utility_stop_network_capture()) { \
      *(ok__) = RTI_TEST_RETCODE_FAILED; \
      RTITestLogger_log( \
              RTI_TEST_LOG_BIT_EXCEPTION, \
              &RTI_TEST_LOG_FAILURE_TEMPLATE, \
              "Error stopping network capture."); \
    } \
    if (!NDDS_Utility_disable_network_capture()) { \
      *(ok__) = RTI_TEST_RETCODE_FAILED; \
      RTITestLogger_log( \
              RTI_TEST_LOG_BIT_EXCEPTION, \
              &RTI_TEST_LOG_FAILURE_TEMPLATE, \
              "Error disabling network capture."); \
    } \
    RTITestSetting_unsetNetworkCaptureSupport(); \
  }

/**************************** Global configuration ****************************/

/** @internal
 * @ingroup RTITestSetting
 * @brief Prototype for all modules setDependencyInstrumentationMask methods.
 */
typedef void (*RTITest_SetDependencyInstrumentationMaskFunction)(
        RTITestLogBitmap verbosity);

/** @internal
 * @ingroup RTITestSetting
 * @brief Global pointer to module setDependencyInstrumentationMask function so
 * that we can modify the logging verbosity from the test.1.0 module.
 */
extern RTITestDllVariable RTITest_SetDependencyInstrumentationMaskFunction
        RTITest_g_setDependencyInstrumentationMask;

/**
 * @ingroup RTITestSetting
 * @brief Prototype for interfering participant detection methods. 
 *  
 * @see RTITest_g_detectInterferingParticipants
 */
typedef int (*RTITest_InterferingParticipantDetectionFunction)(
        int domainId);

/**
 * @ingroup RTITestSetting
 * @brief Global pointer to a function with the signature
 *        RTITest_InterferingParticipantDetectionFunction.
 * @details When the command-line parameter
 *        -enableInterferenceDetection is provided to the
 *         tester, if this pointer is not NULL then the function
 *         it points to will be called in case a test fails. The
 *         Unit Test Framework provides an implementation of
 *         this function,
 *         DDSCTestHelper_detectInterferingParticipants. The
 *         main use case of this function is to try and detect
 *         any other participants that may be running at the
 *         same time as the unit tests and could therefore cause
 *         a test to fail due to interference.
 */
extern RTITestDllVariable RTITest_InterferingParticipantDetectionFunction
        RTITest_g_detectInterferingParticipants;

/* Global variable with the current test name */
extern RTITestDllVariable const char *RTITest_g_currentTestName;

/* Global variable with the current test name if it needs to be allocated */
extern RTITestDllVariable char *RTITest_g_currentTestNameAllocated;

/* Global variable with the lastest assert error message */
extern RTITestDllVariable char *RTITest_g_lastAssertErrorMessage;

/* If a new retcode is added, update this value and RTITest_assertOk */
#define RTI_TEST_NUM_RETCODES 14

extern RTITestDllVariable const char
        *RTITest_g_returnCodeStrings[RTI_TEST_NUM_RETCODES + 1];

/* Max size of RTITest_g_lastAssertErrorMessage buffer */
#define RTI_TEST_MAX_ASSERT_ERROR_MESSAGE_SIZE 1024 * 2

/* Max size of RTITest_g_currentTestNameAllocated buffer */
#define RTI_TEST_MAX_CURRENT_TEST_NAME_ALLOCATED_SIZE 128

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* rtitest_setting_impl_h */
