/**
 * (c) Copyright, Real-Time Innovations, 2021
 * All rights reserved.
 * No duplications, whole or partial, manual or electronic, may be made
 * without express written permission.  Any such copies, or
 * revisions thereof, must display this notice unaltered.
 * This code contains trade secrets of Real-Time Innovations, Inc.
 *
 * Description
 * -----------
 * RTI definitions of cryptography library structures and macros to enable
 * supporting for OpenSSL.
 * A user application does not need to include this header file. It's only
 * intended for internal use by other connextdds modules.
 * =========================================================================
 */

#ifndef security_crypto_lib_adapter_openssl_h
#define security_crypto_lib_adapter_openssl_h

#define RTI_OPENSSL_VERSION_1_0_2 0x10002000L
#define RTI_OPENSSL_VERSION_1_1_0 0x10100000L
#define RTI_OPENSSL_VERSION_3_0_0 0x30000000L

#ifndef RTI_WOLFSSL

  /*
   * This definition must be done before any other openssl include.
   *
   * CRYPTO_dynlock_value is a structure from <openssl/crypto.h> that must be
   * defined by the user. It contains the data needed to handle the locks in the
   * callbacks that OpenSSL uses for handling multi-threading.
   * The definition to RTI_Security_Mutex muts be done before including
   * <openssl/crypto.h>. This affects only OpenSSL versions before 1.1.0 (not
   * used in newer versions).
   */
  #define CRYPTO_dynlock_value RTI_Security_Mutex

  #define RTI_SUPPORT_DH
  #define RTI_SUPPORT_X509_V3_KEY_USAGE_EXTENSION
  #define RTI_SUPPORT_X509_V3_KEY_USAGE_EXTENSION_ENFORCEMENT
  #define RTI_SUPPORT_OPENSSL_ENGINES
  #define RTI_SUPPORT_RSA_PSS_CERTS

  /*
   * This operation casts implicitly from the (void *) pointer returned by
   * EVP_PKEY_get0 to an octet string. We can do this because OpenSSL allocates
   * an octet string for the hmac keys (EVP_PKEY_new_raw_private_key calls
   * set_priv_key, which in the case of hmac keys uses the
   * ASN1_OCTET_STRING_new/set/free functions).
   */
  #define EVP_PKEY_set0(pkey__, data__, dataLength__) \
    ASN1_OCTET_STRING_set(EVP_PKEY_get0(pkey__), (data__), (dataLength__))

  /*
   * This operation resets a BIO to its initial state, before anything was read
   * from it. Returns DDS_BOOLEAN_TRUE on success. See
   * https://www.openssl.org/docs/man3.0/man3/BIO_reset.html
   */
  #define RTI_Security_CryptoLibAdapter_BIO_reset(bio__) \
    (BIO_reset(bio__) == ((BIO_method_type(bio__) == BIO_TYPE_FILE) ? 0 : 1))

  #if OPENSSL_VERSION_NUMBER < RTI_OPENSSL_VERSION_1_1_0
    #include "security_crypto_lib_adapter_openssl102.h"
  #elif OPENSSL_VERSION_NUMBER < RTI_OPENSSL_VERSION_3_0_0
    #include "security_crypto_lib_adapter_openssl111.h"
  #else
    #include "security_crypto_lib_adapter_openssl30.h"
  #endif /* OPENSSL_VERSION_NUMBER < RTI_OPENSSL_VERSION_1_1_0 */

  #ifdef RTI_SUPPORT_OCSP
    #include <openssl/ocsp.h>
  #endif /* RTI_SUPPORT_OCSP */

  /***************************************************************************
   * Signature Identifiers
   **************************************************************************/
  /*
   * Needed because the values returned by X509_get_signature_nid differ between
   * OpenSSL and wolfSSL.
   */
  #define RTI_SECURITY_SIGNATURE_ID_ECDSA_SHA384 NID_ecdsa_with_SHA384
  #define RTI_SECURITY_SIGNATURE_ID_ECDSA_SHA256 NID_ecdsa_with_SHA256
  #define RTI_SECURITY_SIGNATURE_ID_RSA_PSS_SHA256 NID_rsassaPss
  #define RTI_SECURITY_SIGNATURE_ID_RSA_PKCS_SHA256 NID_sha256WithRSAEncryption
#ifdef RTI_SUPPORT_ED25519
  #define RTI_SECURITY_SIGNATURE_ID_EDDSA_ED25519 NID_ED25519
#endif /* RTI_SUPPORT_ED25519 */
#ifdef RTI_SUPPORT_ED448
  #define RTI_SECURITY_SIGNATURE_ID_EDDSA_ED448 NID_ED448
#endif /* RTI_SUPPORT_ED448 */

  /*
   * Unknown Curve Identifier.
   * OpenSSL defines the NIDs with preprocessor macros. Unlike wolfSSL,
   * we don't have static analysis warnings when using NID_undef.
   */
  #define RTI_SECURITY_EC_CURVE_ID_UNKNOWN NID_undef

/*
 * Prevent a warning due to EVP_DigestSignUpdate accepting a size_t count
 * in OpenSSL, but an unsigned int count in wolfSSL.
 */
typedef size_t RTI_Security_CryptoLibAdapterDigestSignUpdateCount;

#define RTI_Security_CryptoLibAdapter_PEM_read_bio_X509_AUX \
    PEM_read_bio_X509_AUX

#define RTI_Security_CryptoLibAdapterX509StoreCtxGetChain \
    X509_STORE_CTX_get0_chain

#define RTI_Security_CryptoLibAdapter_BIO_new_mem_buf BIO_new_mem_buf

#define RTI_Security_CryptoLibAdapter_X509_NAME_print_ex X509_NAME_print_ex

#endif /* RTI_WOLFSSL */
#endif /* security_crypto_lib_adapter_openssl_h */
