/**
 * (c) Copyright, Real-Time Innovations, 2022
 * All rights reserved.
 * No duplications, whole or partial, manual or electronic, may be made
 * without express written permission.  Any such copies, or
 * revisions thereof, must display this notice unaltered.
 * This code contains trade secrets of Real-Time Innovations, Inc.
 *
 * Description
 * -----------
 * RTI definitions of cryptography library structures and macros to enable
 * supporting for OpenSSL 3.0.0 (or greater).
 * A user application does not need to include this header file. It's only
 * intended for internal use by other connextdds modules.
 * =========================================================================
 */

#ifndef security_crypto_lib_adapter_openssl30_h
#define security_crypto_lib_adapter_openssl30_h

#if !defined(RTI_WOLFSSL) \
    && (OPENSSL_VERSION_NUMBER >= RTI_OPENSSL_VERSION_3_0_0)

  /*
   * These definitions must be before the
   * #include "security_crypto_lib_adapter_openssl111.h". Once the sizes are
   * defined, they will not be overwritten.
   */
  #ifndef RTI_SECURITY_PLUGINS_SUPPORT_THREAD_SPECIFIC_STORAGE
    #ifndef RTI_SECURITY_CRYPTO_LIB_VERSION_ADAPTER_EVP_CIPHER_CTX_SIZE
      /*
       * This structure is the same size as EVP_CIPHER_CTX in 64-bit. We use
       * this to declare an EVP_CIPHER_CTX on the stack instead of the heap.
       */
      #define RTI_SECURITY_CRYPTO_LIB_VERSION_ADAPTER_EVP_CIPHER_CTX_SIZE 184
    #endif /* RTI_SECURITY_CRYPTO_LIB_VERSION_ADAPTER_EVP_CIPHER_CTX_SIZE */
    #ifndef RTI_SECURITY_CRYPTO_LIB_VERSION_ADAPTER_EVP_MD_CTX_SIZE
      /*
       * This structure is the same size as EVP_MD_CTX in 64-bit. We use this
       * to declare an EVP_MD_CTX on the stack instead of the heap.
       */
      #define RTI_SECURITY_CRYPTO_LIB_VERSION_ADAPTER_EVP_MD_CTX_SIZE 72
    #endif /* RTI_SECURITY_CRYPTO_LIB_VERSION_ADAPTER_EVP_MD_CTX_SIZE */
  #endif /* RTI_SECURITY_PLUGINS_SUPPORT_THREAD_SPECIFIC_STORAGE */

  #define RTI_Security_CryptoLibAdapterEvpPkey_equals(pkey1__, pkey2__) \
      (EVP_PKEY_eq(pkey1__, pkey2__))

  #include "security_crypto_lib_adapter_openssl111.h"
  #include "openssl/core_names.h"
  #include "openssl/self_test.h"
  #define RTI_SUPPORT_OSSL_PARAM
  #define RTI_SUPPORT_EVP_EC_GEN
  #define RTI_SUPPORT_OCSP

  /*
   * https://www.openssl.org/docs/man3.0/man7/migration_guide.html :
   * "OpenSSL now loads error strings automatically". So the private function
   * ossl_err_get_state_int has replaced the public function ERR_get_state.
   * - For the main thread, ossl_err_get_state_int happens when calling
   *   OPENSSL_init_crypto, which is called internally in the code path of
   *   RAND_seed.
   * - For a thread that needs to decode messages, ossl_err_get_state_int
   *   happens when calling ERR_set_mark, which is called internally in the
   *   code path of EVP_DigestSignInit. When using custom allocators, we call
   *   ERR_set_mark upon per-thread allocator creation to force that allocation
   *   to go to regular memory (as opposed to go to the inline memory
   *   allocator). This is because otherwise, if a user application cleans up
   *   the worker state (i.e., calls
   *   DDS_DomainParticipantFactory_unregister_thread) for a thread that is
   *   part of a thread pool, and then that thread is reused for security, we
   *   will try to reuse the error state, which was already destroyed.
   */
  #define RTI_SUPPORT_AUTO_ALLOCATE_ERROR_STATE

  /*
   * https://www.openssl.org/docs/manmaster/man3/EVP_CIPHER_fetch.html :
   * "EVP_CIPHER_fetch(), EVP_CIPHER_free(), ... were added in 3.0."
   */
  #define RTI_SUPPORT_CIPHER_FETCH

  /*
   * According to
   * https://www.openssl.org/docs/man3.0/man3/ENGINE_load_private_key.html ,
   * engines are deprecated and replaced by the provider API.
   */
  #undef RTI_SUPPORT_OPENSSL_ENGINES
  #define RTI_SUPPORT_OPENSSL_PROVIDERS

    /* Input/output of SSL_CTX_set_options */
    typedef uint64_t RTI_Security_CryptoLibAdapterSslCtxOptions;

  extern RTIBool RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_getGcmTag(
        RTI_Security_CryptoLibAdapterEvpCipherCtxHandle *ctx,
        void *tagPointer, /* output */
        int tagLength /* input */);
  extern RTIBool RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_setGcmTag(
        RTI_Security_CryptoLibAdapterEvpCipherCtxHandle *ctx,
        void *tagPointer, /* input */
        int tagLength /* input */);

#endif /* !defined(RTI_WOLFSSL) && (OPENSSL_VERSION_NUMBER >= RTI_OPENSSL_VERSION_3_0_0) */
#endif /* security_crypto_lib_adapter_openssl30_h */
